////////////////////////////////////////////////////////////////////////
// Class:       TimestampFilter
// Module Type: filter
// File:        TimestampFilter_module.cc
//
// This filter checks the event timestamp (within spill) and requires
// that it be within some user-defined limit.  Useful way to select beam
// events (1.2 - 5.5 sec).
//
// Option to require event contain non-empty container of TPC wire digits.
//
// Generated at Wed Oct 12 15:23:51 2016 by William Foreman using artmod
// from cetpkgsupport v1_10_02.
////////////////////////////////////////////////////////////////////////

// Framework includes
#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art_root_io/TFileService.h"
#include "canvas/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"
#include "messagefacility/MessageLogger/MessageLogger.h"

// LArSoft includes
#include "lardataobj/RawData/OpDetPulse.h"
#include "lardataobj/RawData/RawDigit.h"

// ROOT includes
#include <TH1F.h>
#include <TH2F.h>

// C++ includes
#include <iostream>
#include <memory>

class TimestampFilter;

class TimestampFilter : public art::EDFilter {
public:
  explicit TimestampFilter(fhicl::ParameterSet const & p);
  // The destructor generated by the compiler is fine for classes
  // without bare pointers or other resource use.

  // Plugins should not be copied or assigned.
  TimestampFilter(TimestampFilter const &) = delete;
  TimestampFilter(TimestampFilter &&) = delete;
  TimestampFilter & operator = (TimestampFilter const &) = delete;
  TimestampFilter & operator = (TimestampFilter &&) = delete;

  // Required functions.
  bool filter(art::Event & e) override;

  // Selected optional functions.
  void beginJob() override;
  void endJob() override;
  void reconfigure(fhicl::ParameterSet const & p) ;

private:

  float fT1;
  float fT2;
  bool  fRequireRawDigits;
  std::string fDAQModuleLabel;
  std::string fDAQModuleInstanceName;
  std::string fDataSelection;

  TH1F* hTimestamps;
  TH1F* hTimestamps_pass;
  TH1F* hEvtCount;

  // Time ranges for different data classes
  float range_pedestal[2];
  float range_beam[2];
  float range_cosmic[2];

};


TimestampFilter::TimestampFilter(fhicl::ParameterSet const & p)
: EDFilter(p)
// Initialize member data here.
{
  this->reconfigure(p);

  // ------------------------------
  // Set ranges
  range_pedestal[0] = 0.0;
  range_pedestal[1] = 1.0;
  range_beam[0]     = 1.2;
  range_beam[1]     = 5.2;
  range_cosmic[0]   = 5.5;
  range_cosmic[1]   = 60.;

  // ------------------------------
  // If a data type has been set in the fcl, 
  // override fT1 and fT2
  if( fDataSelection != "" ) {
    std::cout<<"TimestampFilter: "<<fDataSelection<<" data type was provided; T1 and T2 will be overriden.\n";
    if      ( fDataSelection == "pedestal" ) { fT1 = range_pedestal[0]; fT2 = range_pedestal[1];}
    else if ( fDataSelection == "beam"     ) { fT1 = range_beam[0];     fT2 = range_beam[1];    }
    else if ( fDataSelection == "cosmic"   ) { fT1 = range_cosmic[0];   fT2 = range_cosmic[1];  }
    else if ( fDataSelection == "all"      ) { fT1 = 0.;                fT2 = 999.;             }
  }

  // TFile service for saving histograms of timestamps
  art::ServiceHandle<art::TFileService> tfs;
  hTimestamps       = tfs->make<TH1F>("Timestamps",";Time in spill [sec]",300,0,60.);
  hTimestamps_pass  = tfs->make<TH1F>("Timestamps_pass",";Time in spill [sec]",300,0.,60.);
  
  // Data type counter
  hEvtCount = tfs->make<TH1F>("EvtCount","Counted events",3,0,3);
  hEvtCount->SetOption("HIST TEXT");
  hEvtCount->GetXaxis()->SetBinLabel(1,"Pedestal");
  hEvtCount->GetXaxis()->SetBinLabel(2,"Beam");      
  hEvtCount->GetXaxis()->SetBinLabel(3,"Cosmic");         
  
}

bool TimestampFilter::filter(art::Event & e)
{
  
  // if MC, then timestamp is meaningless
  if( !e.isRealData() ) return true;
  
  // -----------------------------------------------
  // Require event contain non-empty container of RawDigits (wires)
  // If empty, skip event immediately.
  art::Handle< std::vector<raw::RawDigit> > DigitHandle;;
  std::vector<art::Ptr<raw::RawDigit> > digit;
  if(e.getByLabel("daq",DigitHandle))
    {art::fill_ptr_vector(digit, DigitHandle);} 
  if( fRequireRawDigits && digit.size() == 0 ) return false;
  
  // Get the timestamp (within the spill cycle) from the opdetpulse
  // objects because I don't know where else this info is saved!
  art::Handle< std::vector< raw::OpDetPulse >> opdetHandle;
  e.getByLabel(fDAQModuleLabel, fDAQModuleInstanceName, opdetHandle);
  
  float timeStamp = -1.;
  
  if( (size_t)opdetHandle->size() > 0 ){
    // All we want is the timestamp so just grab the first opdetpulse
    // we can find and call it a day
    art::Ptr< raw::OpDetPulse > ThePulsePtr(opdetHandle,0); 
//    raw::OpDetPulse pulse = *ThePulsePtr;
    timeStamp = ((float)(*ThePulsePtr).PMTFrame()*8.)/1.0e09;
    //std::cout<<"Timestamp = "<<timeStamp<<" sec\n";
    hTimestamps->Fill(timeStamp);

    // Counting up "beam" and "cosmic" events
    if(       timeStamp >= range_pedestal[0] && timeStamp < range_pedestal[1] ) hEvtCount->Fill(0);
    else if(  timeStamp >= range_beam[0] && timeStamp < range_beam[1]         ) hEvtCount->Fill(1);
    else if(  timeStamp >= range_cosmic[0] && timeStamp < range_cosmic[1]     ) hEvtCount->Fill(2);
  }
  
  if( timeStamp >= fT1 && timeStamp <= fT2 ) {
    hTimestamps_pass->Fill(timeStamp);
    return true;
  }
  
  return false;

}

void TimestampFilter::beginJob()
{
}


void TimestampFilter::endJob()
{
  std::cout
  <<"====================================\n"
  <<"=====     TimestampFilter   ========\n\n"
  <<"Total events         : "<<(int)hTimestamps->GetEntries()<<"\n"
  <<" -pedestal (0-1 sec) : "<<(int)hEvtCount->GetBinContent(1)<<"\n"
  <<" -beam (1.2-5.2 sec) : "<<(int)hEvtCount->GetBinContent(2)<<"\n"
  <<" -cosmic (> 5.2 sec) : "<<(int)hEvtCount->GetBinContent(3)<<"\n\n"
  <<"Allowed range        : "<<fT1<<" - "<<fT2<<"   "<<fDataSelection<<"\n"
  <<"Events passed        : "<<(int)hTimestamps_pass->GetEntries()<<"\n\n"
  <<"====================================\n";
}


void TimestampFilter::reconfigure(fhicl::ParameterSet const & p)
{
  fT1				= p.get< float  >("T1",0);
  fT2				= p.get< float  >("T2",60);
  fDAQModuleLabel               = p.get< std::string >  ("DAQModule","daq");
  fDAQModuleInstanceName        = p.get< std::string >  ("DAQInstanceName","");
  fRequireRawDigits             = p.get< bool >         ("RequireRawDigits",true); 
  fDataSelection                = p.get< std::string >  ("DataSelection",""); 
}


DEFINE_ART_MODULE(TimestampFilter)
